
###### Definition of type Cvect: generates the structure for cost vectors ############################################################################
type Cvect
    sense::Int # +1 is for minimization and -1 is for maximization 
    rxns::Vector{Int} # Reaction indices that appear in cost vector
    wts::Vector{Float64} # Coefficients of the reaction indices 
    comment::String # String indicating the nature of the cost vector
    Cvect(sense,rxns,wts,comment) = length(rxns) == length(wts) ? new(sense,rxns,wts,comment) : error("Inconsistent size of rxns and wts vectors.");
end
# Constructors Cvect
Cvect(sense::Int,rxns::Int,wts::Int) = Cvect(sense,[rxns],[convert(Float64,wts)],"");
Cvect(sense::Int,rxns::Int,wts::Float64) = Cvect(sense,[rxns],[wts],"");
Cvect(sense::Int,rxns::Int,wts::Int,comment::String) = Cvect(sense,[rxns],[convert(Float64,wts)],comment);
Cvect(sense::Int,rxns::Int,wts::Float64,comment::String) = Cvect(sense,[rxns],[wts],comment);
Cvect() = Cvect(0,[0],[0],"");


###### Definition of type FBAmodel: generates the structure to store relevant information of FBA model ###################################################
type FBAmodel
    S::SparseMatrixCSC{Float64,Int64} # Stoichiometry matrix
    b::Vector{Float64} # Right-hand side: usually all zeroes
    lb::Vector{Float64} # Lower bound vector
    ub::Vector{Float64} # Upper bound vector
    C::Vector{Cvect} # Cost vector information
    DB::Float64 # Default bound equivalent to infinity. It's default value will be 1000.
    Penflag::Bool # To use a more specialized penalty function
    Penvector::Vector{Int} # Indices of reactions that will concentrate the penalty
    Pentype::Vector{Int} # For each reaction in penvector, we need to select 0 for lower bound, 1 for upper bound, 2 for both
    comment::String # String for comments
    FBAmodel(S,b,lb,ub,C,DB,Penflag,Penvector,Pentype,comment) = length(Penvector) == length(Pentype) ? new(S,b,lb,ub,C,DB,Penflag,Penvector,Pentype,comment) : error("Penvector and Pentype should have the same size.");
end

# Constructor FBAmodel
FBAmodel() = FBAmodel(sparse([0.0 0.0]),[0.0],[0.0],[0.0],[Cvect()],1000.0,false,[0],[0],"");


##### Function that extracts FBA model from mat file and returns it as an FBAmodel #######################################################################
function extract(var::Dict)
    model = FBAmodel();
    model.S = copy(var["S"]);
    model.b = copy(vec(var["b"]));
    model.lb = copy(vec(var["lb"]));
    model.ub = copy(vec(var["ub"]));
    return model
end

##### Function that replaces DB for Inf and returns full row rank system #################################################################################
function FullRank(model::FBAmodel)
    DB = model.DB;
    ind = find(x->(x >= DB),model.ub);
    model.ub[ind] = Inf;
    ind = find(x->(x <= -DB),model.lb);
    model.lb[ind] = -Inf;

    # Reformulating Ax = b for an equivalent system using QR factorization.
    # AE = QR, E-1 = E', Q-1 = Q'
    # QRE'x = b
    # RE'x = Q'b
    # Since R is an upper triangular matrix, only the first r rows of R (where
    # r is the rank of R) will contain the relevant information.
    # S = full(model.S);
    # F = qrfact(S,Val{true}); 
    # r = rank(S);
    # Q = F[:Q];
    # R = F[:R];
    # P = F[:P];
    F = qrfact(model.S); 
    Sfull = full(model.S);
    r = rank(Sfull);
    bnew = Base.SparseArrays.SPQR.qmult(Base.SparseArrays.SPQR.QTX, F, Base.SparseArrays.CHOLMOD.Dense(model.b));
    Anew = sparse(Base.SparseArrays.SPQR.qmult(Base.SparseArrays.SPQR.QTX, F, Base.SparseArrays.CHOLMOD.Dense(Sfull)));
    r1 = rank(full(Anew));
    if r1 != r
       error("The full rank code is changing the original rank of the matrix");
    end
    model.S = Anew[1:r,:];
    model.b = bnew[1:r];
   # Base.SparseArrays.SPQR.free!(F);
end

using MAT

home = "/home/jagomezr/Desktop/SPQR failure/"

name = "iND750";
file = matopen((string(home,name,".mat")))
var = read(file, name) 
close(file)

iND750 = extract(var);

name = "iRC1080";
file = matopen((string(home,name,".mat")))
var = read(file, name)
close(file)

iRC1080 = extract(var);

# Define DB: any value >= DB will be set equal to Infinity
iND750.DB = 1000;
iRC1080.DB = 1000;

# Transform models into full row rank models. This will allow to reuse SBML models without going through 
# the rank calculation process multiple times. 
FullRank(iND750);
FullRank(iRC1080);

# Declare how many different models we will have. Here, we create copies of a model, if a single model is to be
# reused. 
nmodel = 2; 
# Pack models
model = Vector{FBAmodel}(nmodel)
model[1] = iND750;
model[2] = iRC1080;


# Define cost vectors
minim = 1;
maxim = -1;

numcostvectors = Vector{Int}(nmodel)
numcostvectors[1] = 5;
numcostvectors[2] = 4;

for i = 1:nmodel
    model[i].C = Vector{Cvect}(numcostvectors[i])
end

# Cost vectors yeast
model[1].C[1]=Cvect(maxim,1266,1,"Maximize growth") 
model[1].C[2]=Cvect(maxim,428,1,"Minimize glucose consumption") 
model[1].C[3]=Cvect(maxim,458,1,"Minimize O2 consumption") 
model[1].C[4]=Cvect(maxim,407,1,"Maximize CO2 production") 
model[1].C[5]=Cvect(maxim,420,1,"Maximize ethanol production") 

# Cost vectors algae
model[2].C[1]=Cvect(maxim,63,1,"Maximize growth") 
model[2].C[2]=Cvect(minim,28,1,"Maximize glucose consumption") 
model[2].C[3]=Cvect(maxim,[24,81],[1,1],"Minimize consumption, maximize production O2") 
model[2].C[4]=Cvect(maxim,407,1, "Minimize CO2 production")
